;;; basic-datatypes.scm


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Fundamental Datatypes ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; This create type synonyms.
;;; Most types are really just card8, card16, or card32 numbers.
;;; It's easier to think of them as `pixmaps' or whatever, and
;;; have functions that operate with those names.

(define-syntax define-simple-type
  (lambda (form rename compare?)
    (let ((symbol-append (lambda syms
			   (string->symbol
			    (apply string-append (map symbol->string syms))))))
      (let ((name (cadr form))
	    (type (caddr form)))
	(call-with-values (lambda ()
			    (cond ((compare? type (rename 'card8))
				   (values (rename 'number->card8)
					   (rename 'number->card8/fill!)
					   (rename 'read-card8)
					   1))
				  ((compare? type (rename 'card16))
				   (values (rename 'number->card16)
					   (rename 'number->card16/fill!)
					   (rename 'read-card16)
					   2))
				  ((compare? type (rename 'card32))
				   (values (rename 'number->card32)
					   (rename 'number->card32/fill!)
					   (rename 'read-card32)
					   4))
				  ;; I don't expect this to execute, just show what's wrong.
				  (else (list 'syntax-error 'define-simple-type
					      "unrecognized base type" type))))
	  (lambda (number->it fill! read-it size)
	    (let ((%begin  (rename 'begin))
		  (%define (rename 'define))
		  (%set!   (rename 'set!))
		  (%cons   (rename 'cons))
		  (%quote  (rename 'quote))
		  (%synonyms (rename '*synonyms*)))
	      (list %begin
		    (list %define (symbol-append 'number-> name) number->it)
		    (list %define (symbol-append 'number-> name '/fill!) fill!)
		    (list %define (symbol-append 'read- name) read-it)
		    ))))))))

(define-simple-type atom         card32)
(define-simple-type bit-gravity  card8)
(define-simple-type bitmask      card32)
(define-simple-type button       card8)
(define-simple-type class        card8)
(define-simple-type colormap     card32)
(define-simple-type cursor       card32)
(define-simple-type depth        card8)
(define-simple-type drawable     card32)
(define-simple-type font         card32)
(define-simple-type fontable     card32)
(define-simple-type gcontext     card32)
(define-simple-type keycode      card8)
(define-simple-type keysym       card32)
(define-simple-type pixmap       card32)
(define-simple-type timestamp    card32)
(define-simple-type visual-id    card32)
(define-simple-type window       card32)


;;; This is, in a sense, the inverse of substring!.
;;; It copies the source string into part of the target-string.

(define fill-string!
  (lambda (target-string offset source-string)
    (do ((source-i 0 (+ source-i 1))
	 (target-i offset (+ target-i 1)))
	((>= source-i (string-length source-string)) #t)
      (if (>= target-i (string-length target-string))
	  (error "offset exceeded target-string length" target-string target-i)
	  (string-set! target-string target-i (string-ref source-string source-i))))))


;;;;;;;;;;;;;;;;;;;;
;;; Abstractions ;;;
;;;;;;;;;;;;;;;;;;;;

;;; (ignore number-of-bytes-to-ignore port)  => unspecified

(define ignore read-string8)


;;; (read-list-of-bytes port number-of-bytes)  => string

(define read-list-of-bytes read-string8)


;;; (symbol->value alist not-found-error)  => (symbol -> value)

(define symbol->value
  (lambda (alist not-found-error)
    (lambda (sym)
      (cond ((assq sym alist) => cdr)
	    (else (not-found-error sym))))))


;;; (value->symbol alist not-found-error)  => (value -> symbol)

(define value->symbol
  (lambda (alist not-found-error)
    (lambda (val)
      (let loop ((alist alist))
	(if (null? alist)
	    (not-found-error val)
	    (let ((top (car alist)))
	      (let ((symbol (car top))
		    (value (cdr top)))
		(if (= val value)
		    symbol
		    (loop (cdr alist))))))))))


;;;;;;;;;;;;;;;;
;;; Booleans ;;;
;;;;;;;;;;;;;;;;

(define read-bool
  (lambda (xserver)
    (let ((bool (read-card8 xserver)))
      (case bool
	((0) #f)
	((1) #t)
	(else (error "input not bool" bool))))))

(define boolean->bool/fill!
  (lambda (str offset boolean)
    (cond ((eq? boolean #t)
	   (number->card8/fill! str offset 1))
	  ((eq? boolean #f)
	   (number->card8/fill! str offset 0))
	  (else (error "expected boolean" boolean)))))
	  
(define boolean->bool
  (lambda (boolean)
    (cond ((eq? boolean #t)
	   (number->card8 1))
	  ((eq? boolean #f)
	   (number->card8 0))
	  (else (error "expected boolean" boolean)))))


;;;;;;;;;;;;;;;;;;;
;;; Bit Gravity ;;;
;;;;;;;;;;;;;;;;;;;

(define bit-gravity-alist `((forget     .  0)
			    (north-west .  1)
			    (north      .  2)
			    (north-east .  3)
			    (west       .  4)
			    (center     .  5)
			    (east       .  6)
			    (south-west .  7)
			    (south      .  8)
			    (south-east .  9)
			    (static     . 10)))


;;; (symbol->bit-gravity symbol)  => bit-gravity

(define symbol->bit-gravity
  (symbol->value bit-gravity-alist
		 (lambda (sym)
		   (error "unknown symbol" sym))))


;;; (bit-gravity->symbol bit-gravity) => symbol

(define bit-gravity->symbol
  (value->symbol bit-gravity-alist
		 (lambda (val)
		   (error "value out of range" val))))

(define read-bit-gravity
  (lambda (xserver)
    (bit-gravity->symbol (read-card8 xserver))))


;;;;;;;;;;;;;;;;;;;
;;; Win Gravity ;;;
;;;;;;;;;;;;;;;;;;;

(define win-gravity-alist `((unmap      .  0)
			    (north-west .  1)
			    (north      .  2)
			    (north-east .  3)
			    (west       .  4)
			    (center     .  5)
			    (east       .  6)
			    (south-west .  7)
			    (south      .  8)
			    (south-east .  9)
			    (static     . 10)))


;;; (symbol->win-gravity symbol) => win-gravity

(define symbol->win-gravity
  (symbol->value win-gravity-alist
		 (lambda (sym)
		   (error "unknown symbol" sym))))


;;; (win-gravity->symbol win-gravity) => symbol

(define win-gravity->symbol
  (value->symbol win-gravity-alist
		 (lambda (val)
		   (error "value out of range" val))))

(define read-win-gravity
  (lambda (xserver)
    (win-gravity->symbol (read-card8 xserver))))


;;;;;;;;;;;;;;;;;;;;
;;; Set-of-Event ;;;
;;;;;;;;;;;;;;;;;;;;

(define event-set-type
  (make-set-type
   32
   0 0 0 0 0 0 0
   'owner-grab-button
   'colormap-change 'property-change 'focus-change 'substructure-redirect
   'substructure-notify 'resize-redirect 'structure-notify 'visibility-change
   'exposure 'keymap-state 'button-motion 'button5-motion
   'button4-motion 'button3-motion 'button2-motion 'button1-motion
   'pointer-motion-hint 'pointer-motion 'leave-window 'enter-window
   'button-release 'button-press 'key-release 'key-press
   ))

(define symbols->set-of-event/fill!
  (lambda (str offset symbols)
    (let ((set (make-set event-set-type 0)))
      (for-each (lambda (symbol)
		  (set-bit set symbol))
		symbols)
      (number->card32/fill! str offset (set->number set)))))

(define read-set-of-event
  (lambda (xserver)
    (set->symbols
     (make-set event-set-type (read-card32 xserver)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; set-of-pointer-event ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define pointer-event-set-type
  (make-set-type
   32
   0 0 0 0 0 0 0 0
   0 0 0 0 0 0 0 0
   0 'keymap-state 'button-motion 'button5-motion
   'button4-motion 'button3-motion 'button2-motion 'button1-motion
   'pointer-motion-hint 'pointer-motion 'leave-window 'enter-window
   'button-release 'button-press 0 0
   ))

(define symbols->set-of-pointer-event/fill!
  (lambda (str offset symbols)
    (let ((set (make-set pointer-event-set-type 0)))
      (for-each (lambda (symbol)
		  (set-bit set symbol))
		symbols)
      (number->card32/fill! str offset (set->number set)))))

(define read-set-of-pointer-event
  (lambda (xserver)
    (set->symbols
     (make-set pointer-event-set-type (read-card32 xserver)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; set-of-device-event ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define device-event-set-type
  (make-set-type
   32
   0 0 0 0 0 0 0 0
   0 0 0 0 0 0 0 0
   0 0 'button-motion 'button5-motion
   'button4-motion 'button3-motion 'button2-motion 'button1-motion
   0 'pointer-motion 0 0
   'button-release 'button-press 'key-release 'key-press
   ))

(define symbols->set-of-device-event/fill!
  (lambda (str offset symbols)
    (let ((set (make-set device-event-set-type 0)))
      (for-each (lambda (symbol)
		  (set-bit set symbol))
		symbols)
      (number->card32/fill! str offset (set->number set)))))

(define read-set-of-device-event
  (lambda (xserver)
    (set->symbols
     (make-set device-event-set-type (read-card32 xserver)))))


;;;;;;;;;;;;;;;;;;;;;;;;;
;;; set-of-keybutmask ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;

(define keybutmask-set-type
  (make-set-type
   16
   0 0 0 'button5 'button4 'button3 'button2 'button1
   'mod5 'mod4 'mod3 'mod2 'mod1 'control 'lock 'shift
   ))

(define read-keybutmask
  (lambda (xserver)
    (let ((n (read-card16 xserver)))
      (set->symbols
       (make-set keybutmask-set-type n)))))

;; (define symbols->set-of-keybutmask/fill!
;;   (lambda (str offset symbols)
;;     (let ((set (make-set keybutmask-set-type 0)))
;;       (for-each (lambda (symbol)
;;                   (set-bit set symbol))
;;                 symbols)
;;       (card16->number/fill! str offset (set->number set)))))


;;;;;;;;;;;;;;;;;;;;;;
;;; set-of-keymask ;;;
;;;;;;;;;;;;;;;;;;;;;;

(define keybutmask-set-type
  (make-set-type
   16
   0 0 0 0 0 0 0 0
   'mod5 'mod4 'mod3 'mod2 'mod1 'control 'lock 'shift
   ))

(define read-keymask
  (lambda (xserver)
    (let ((n (read-card16 xserver)))
      (set->symbols
       (make-set keybutmask-set-type n)))))

;; (define symbols->set-of-keymask/fill!
;;   (lambda (str offset symbols)
;;     (let ((set (make-set keybutmask-set-type 0)))
;;       (for-each (lambda (symbol)
;;                   (set-bit set symbol))
;;                 symbols)
;;       (card16->number/fill! str offset (set->number set)))))


;;;;;;;;;;;;
;;; Arcs ;;;
;;;;;;;;;;;;

(define arc-size 12)

(define bad-arc-message
  (string-append
   "bad arc format: expected either "
   "'(x y width height angle1 angle2) or "
   "'#(x y width height angle1 angle2)"))

(define arc-structures->arcs
  (lambda (list-of-arcs)
    (let* ((number-of-arcs (length list-of-arcs))
	   (str (make-string (* number-of-arcs arc-size))))
      (let loop ((i 0) (list list-of-arcs))
	(if (null? list)
	    str
	    (let ((arc (car list)))
	      (cond ((pair? arc)   (arc-list/fill!   str i arc))
		    ((vector? arc) (arc-vector/fill! str i arc))
		    (else (error bad-arc-message arc)))
	      (loop (+ i arc-size) (cdr list))))))))

(define arc-list/fill!
  (lambda (str offset arc-list)
    (if (not (= (length arc-list) 6))
	(error bad-arc-message arc-list)
	(begin
	  (number->int16/fill!  str     offset     (first  arc-list))
	  (number->int16/fill!  str  (+ offset  2) (second arc-list))
	  (number->card16/fill! str  (+ offset  4) (third  arc-list))
	  (number->card16/fill! str  (+ offset  6) (fourth arc-list))
	  (number->int16/fill!  str  (+ offset  8) (fifth  arc-list))
	  (number->int16/fill!  str  (+ offset 10) (sixth  arc-list))
	  ))))

(define arc-vector/fill!
  (lambda (str offset arc-vec)
    (if (not (= (vector-length arc-vec) 6))
	(error bad-arc-message arc-vec)
	(begin
	  (number->int16/fill!  str     offset     (vector-ref arc-vec 0))
	  (number->int16/fill!  str  (+ offset  2) (vector-ref arc-vec 1))
	  (number->card16/fill! str  (+ offset  4) (vector-ref arc-vec 2))
	  (number->card16/fill! str  (+ offset  6) (vector-ref arc-vec 3))
	  (number->int16/fill!  str  (+ offset  8) (vector-ref arc-vec 4))
	  (number->int16/fill!  str  (+ offset 10) (vector-ref arc-vec 5))
	  ))))


;;;;;;;;;;;;;;
;;; Points ;;;
;;;;;;;;;;;;;;

;;; A point may be either of #(x y), '(x . y), or '(x y)

(define point-size 4)

(define bad-point-message
  "bad point format:  expected '(x . y), '(x y), or '#(x y)")

(define point-structures->points
  (lambda (points)
    (let* ((number-of-points (length points))
	   (str (make-string (* number-of-points point-size))))
      (let loop ((i 0) (list points))
	(if (null? list)
	    str
	    (let ((point (car list)))
	      (cond ((pair? point)   (point-list/fill! str i point))
		    ((vector? point) (point-vec/fill!  str i point))
		    (else (error bad-point-message point)))
	      (loop (+ i point-size) (cdr list))))))))

(define point-list/fill!
  (lambda (str offset point)
    (cond ((not (pair? (cdr point)))
	   (number->int16/fill! str    offset    (car point))
	   (number->int16/fill! str (+ offset 2) (cdr point)))
	  (else
	   (number->int16/fill! str    offset    (car point))
	   (number->int16/fill! str (+ offset 2) (cadr point)))
	  )))

(define point-vec/fill!
  (lambda (str offset point)
    (if (not (= (vector-length point) 2))
	(error bad-point-message point)
	(begin
	  (number->int16/fill! str    offset    (vector-ref point 0))
	  (number->int16/fill! str (+ offset 2) (vector-ref point 1))
	  ))))


;;;;;;;;;;;;;;;;;;
;;; Rectangles ;;;
;;;;;;;;;;;;;;;;;;

(define rectangle-size 8)

(define bad-rectangle-message
  "bad rectangle format: expected either '(x y width height) or '#(x y width height)")

(define rectangle-structures->rectangles
  (lambda (rectangle-structures)
    (let* ((number-of-rectangles (length rectangle-structures))
	   (str (make-string (* number-of-rectangles rectangle-size))))
      (let loop ((i 0) (recs rectangle-structures))
	(if (null? recs)
	    str
	    (let ((rectangle (car recs)))
	      (cond ((pair? rectangle)
		     (rectangle-list/fill! str i rectangle))
		    ((vector? rectangle)
		     (rectangle-vector/fill! str i rectangle))
		    (else (error bad-rectangle-message rectangle)))
	      (loop (+ i rectangle-size) (cdr recs))))))))

(define rectangle-list/fill!
  (lambda (str offset rectangle-list)
    (if (not (= (length rectangle-list) 4))
	(error bad-rectangle-message rectangle-list)
	(begin
	  (number->int16/fill!  str    offset    (first  rectangle-list))
	  (number->int16/fill!  str (+ offset 2) (second rectangle-list))
	  (number->card16/fill! str (+ offset 4) (third  rectangle-list))
	  (number->card16/fill! str (+ offset 6) (fourth rectangle-list))
	  ))))

(define rectangle-vector/fill!
  (lambda (str offset vec)
    (if (and (not (vector? vec))
	     (= 4 (vector-length vec)))
	(error bad-rectangle-message vec)
	(begin
	  (number->int16/fill!  str    offset    (vector-ref vec 0))
	  (number->int16/fill!  str (+ offset 2) (vector-ref vec 1))
	  (number->card16/fill! str (+ offset 4) (vector-ref vec 2))
	  (number->card16/fill! str (+ offset 6) (vector-ref vec 3))
	  ))))


;;;;;;;;;;;;;
;;; Hosts ;;;
;;;;;;;;;;;;;

;; scshy.

(define internet-address->byte-sequence
  (lambda (address)
    (let* ((internet-address-regex "([0-9]+)[.]([0-9]+)[.]([0-9]+)[.]([0-9]+)")
	   (match (string-match internet-address-regex address)))
      (if (not match)
	  (error "can't parse internet address; expecting `<num>.<num>.<num>.<num>'" address)
	  (let* ((first-char (lambda (str) (string-ref str 0)))
		 (match_i    (lambda (i) (match:substring match i)))
		 (string-number->byte (lambda (s)
					(first-char (number->card8 (string->number s))))))
	    (do ((str (make-string 4))
		 (i 0 (+ i 1)))
		((>= i 4) str)
	      (string-set! str i (string-number->byte (match_i (+ i 1))))))))))
