;;; events.scm


(define redirect-event
  (lambda (xserver code)
    (case code
      ((2)  (read-key-press-event xserver))
      ((3)  (read-key-release-event xserver))
      ((4)  (read-button-press-event xserver))
      ((5)  (read-button-release-event xserver))
      ((6)  (error "MotionNotify (6) unimplemented"))
      ((7)  (read-enter-notify-event xserver))
      ((8)  (read-leave-notify-event xserver))
      ((9)  (error "FocusIn (9) unimplemented"))
      ((10) (error "FocusOut (10) unimplemented"))
      ((11) (error "KeymapNotify (11) unimplemented"))
      ((12) (read-expose-event xserver))
      ((13) (read-graphics-exposure-event xserver))
      ((14) (read-no-expose-event xserver))
      ((15) (error "VisibilityNotify (15) unimplemented"))
      ((16) (error "CreateNotify (16) unimplemented"))
      ((17) (error "DestroyNotify (17) unimplemented"))
      ((18) (error "UnmapNotify (18) unimplemented"))
      ((19) (error "MapNotify (19) unimplemented"))
      ((20) (error "MapRequest (20) unimplemented"))
      ((21) (error "ReparentNotify (21) unimplemented"))
      ((22) (error "ConfigureNotify (22) unimplemented"))
      ((23) (error "ConfigureRequest (23) unimplemented"))
      ((24) (error "GravityNotify (24) unimplemented"))
      ((25) (error "ResizeRequest (25) unimplemented"))
      ((26) (error "CirculateNotify (26) unimplemented"))
      ((27) (error "CirculateRequest (27) unimplemented"))
      ((28) (error "PropertyNotify (28) unimplemented"))
      ((29) (error "SelectionClear (29) unimplemented"))
      ((30) (error "SelectionRequest (30) unimplemented"))
      ((31) (error "SelectionNotify (31) unimplemented"))
      ((32) (error "ColormapNotify (32) unimplemented"))
      ((33) (error "ClientMessage (33) unimplemented"))
      ((34) (error "MappingNotify (34) unimplemented"))
      (else (error "unsupported event" code)))))

(define display-event
  (lambda (event)
    (cond ((button-press-event? event)      (display-button-press-event event))
	  ((button-release-event? event)    (display-button-release-event event))
	  ((key-press-event? event)         (display-key-press-event event))
	  ((key-release-event? event)       (display-key-release-event event))
	  ((expose-event? event)            (display-expose-event event))
	  ((enter-notify-event? event)      (display-enter-notify-event event))
	  ((leave-notify-event? event)      (display-leave-notify-event event))
	  ((graphics-exposure-event? event) (display-graphics-exposure-event event))
	  ((no-expose-event? event)         (display-no-expose-event event))
	  (else (error "unsupported event" event)))))

;;;
;;; Macro to build a record, read function, and display function for an event
;;;

;;; The define-event macro:
;;;
;;; (define-event <event-name> (<entry>*))
;;;
;;; <event-name> = <symbol>
;;; For example, `button-press-event'.
;;;
;;; <entry> = (<field> <type>)
;;;         = (<field> <type> <alter-function>)
;;;         = ((<field1> <field2>) <type> <alter-function>)
;;; If unspecified, <alter-function> defaults to the identity function.
;;; If two fields are given, <alter-function> must return two values.
;;;
;;; <field> = <symbol>
;;;         = unused
;;; For example, `x'.
;;; `unused' is a special case, where the value needs to be read, but not recorded.
;;;
;;; <type>  = <symbol>
;;;         = (f <args>*)
;;; Examples: `card8', `(ignore 20)'.
;;; If <type> = (f <args>*), the reader becomes (f <args>* xserver).


;;; Should just use explict-renaming macros for all of it.

(define-syntax evaluating-define
  (lambda (form rename compare)
    (let ((name (eval (cadr form) (interaction-environment)))  ;; absurd
	  (body (caddr form)))
      `(define ,name ,body))))

(define-syntax symbol-call
  (lambda (form rename compare)
    (let ((f    (cadr form))
	  (args (cddr form)))
      (let ((func (eval f (interaction-environment))))
	`(,func ,@args)))))

(define-syntax symbol-apply
  (lambda (form rename compare)
    (let ((f    (cadr form))
	  (args (caddr form)))
      (let ((func (eval f (interaction-environment))))
	`(,(rename 'apply) ,func ,args)))))


(define-syntax define-event-record
  (syntax-rules (unused)
    ((define-event-record name fields)
     (define-event-record name fields ()))
    ((define-event-record name () (fields ...))
     (define-record name fields ...))
    ((define-event-record name ((field type) more ...) seen)
     (define-event-record name ((field type identity) more ...) seen))
    ((define-event-record name ((unused type alt) more ...) (seen-fields ...))
     (define-event-record name (more ...) (seen-fields ...)))
    ((define-event-record name (((value1 value2) type alt) more ...) (seen-fields ...))
     (define-event-record name (more ...) (seen-fields ... value2 value1)))
    ((define-event-record name ((field type alt) more ...) (seen-fields ...))
     (define-event-record name (more ...) (seen-fields ... field)))
    ))

(define-syntax type-reader
  (syntax-rules ()
    ((type-reader (arg ...) xserver)
     (arg ... xserver))
    ((type-reader x xserver)
     (symbol-call (symbol-append 'read- (quote x)) xserver))
    ))

(define-syntax event-reader-let
  (syntax-rules (unused)
    ((event-reader-let name xserver () (fields ...))
     (symbol-call (symbol-append 'make- (quote name)) fields ...))
    ((event-reader-let name xserver ((field type) more ...) fields)
     (event-reader-let name xserver ((field type identity) more ...) fields))
    ((event-reader-let name xserver (((value1 value2) type alter-func) more ...) (fields ...))
     (call-with-values (lambda () (alter-func (type-reader type xserver)))
       (lambda (value1 value2)
	 (event-reader-let name xserver (more ...) (fields ... value1 value2)))))
    ((event-reader-let name xserver ((unused type alter-func) more ...) (fields ...))
     (let ((unused (type-reader type xserver)))
       (event-reader-let name xserver (more ...) (fields ...))))
    ((event-reader-let name xserver ((field type alter-func) more ...) (fields ...))
     (let ((field (alter-func (type-reader type xserver))))
       (event-reader-let name xserver (more ...) (fields ... field))))
    ))

(define-syntax define-event-reader
  (syntax-rules ()
    ((define-event-reader name lines)
     (evaluating-define (symbol-append 'read- (quote name))
       (lambda (xserver)
	 (format #t "Reading ~a!~%" (quote name))
	 (event-reader-let name xserver lines ()))))
    ))

(define-syntax event-format
  (syntax-rules (unused)
    ((event-format name event () (strings ...) (accessors ...))
     (format #t (string-append "~a~%" strings ...) (quote name) accessors ...))
    ((event-format name event ((field type) more ...) strings accessors)
     (event-format name event ((field type identity) more ...) strings accessors))
    ((event-format name event ((unused type alt) more ...) (strings ...) (accessors ...))
     (event-format name event (more ...) (strings ...) (accessors ...)))
    ((event-format name event (((value1 value2) type alter-func) more ...)
		   (strings ...) (accessors ...))
     (event-format name event
		   (more ...)
		   (strings ...
			    "  " (symbol->string (quote value1)) ":  ~a~%"
			    "  " (symbol->string (quote value2)) ":  ~a~%")
		   (accessors ... 
			      (symbol-apply (symbol-append (quote name) (quote :) (quote value1))
					    (list event))
			      (symbol-apply (symbol-append (quote name) (quote :) (quote value2))
					    (list event)))))
    ((event-format name event ((field type alt) more ...) (strings ...) (accessors ...))
     (event-format
      name event
      (more ...)
      (strings ... "  " (symbol->string (quote field)) ":  ~a~%")
      (accessors ... (symbol-apply (symbol-append (quote name) (quote :) (quote field))
				   (list event)))))
    ))

(define-syntax define-event-formatter
  (syntax-rules ()
    ((define-event-formatter name lines)
     (evaluating-define (symbol-append 'display- (quote name))
       (lambda (event)
	 (event-format name event lines () ()))))
    ))

(define-syntax define-event
  (syntax-rules ()
    ((define-event name lines)
     (begin
       (define-event-record name lines)
       (define-event-reader name lines)
       (define-event-formatter name lines)))
    ))


;;; This one is used a lot.

(define zero-window
  (lambda (window)
    (if (zero? window)
	'none
	window)))


;;;;;;;;;;;;;;;;;;;;
;;; button-press ;;;
;;;;;;;;;;;;;;;;;;;;

(define-event button-press-event
  ((code             card8)
   (detail           button)
   (sequence-number  card16)
   (time             timestamp)
   (root             window)
   (event            window)
   (child            window      zero-window)
   (root-x           int16)
   (root-y           int16)
   (event-x          int16)
   (event-y          int16)
   (state            keybutmask)
   (same-screen      bool)
   (unused           (ignore 1))
   ))


;;;;;;;;;;;;;;;;;;;;;;
;;; button-release ;;;
;;;;;;;;;;;;;;;;;;;;;;

(define-event button-release-event
  ((code             card8)
   (detail           button)
   (sequence-number  card16)
   (time             timestamp)
   (root             window)
   (event            window)
   (child            window      zero-window)
   (root-x           int16)
   (root-y           int16)
   (event-x          int16)
   (event-y          int16)
   (state            keybutmask)
   (same-screen      bool)
   (unused           (ignore 1))
   ))


;;;;;;;;;;;;;;;;;;;;
;;; enter-notify ;;;
;;;;;;;;;;;;;;;;;;;;

(define enter-notify-detail
  (lambda (detail)
    (case detail
      ((0) 'ancestor)
      ((1) 'virtual)
      ((2) 'inferior)
      ((3) 'nonlinear)
      ((4) 'nonlinear-virtual)
      (else (error "Bad detail received from xserver" detail)))))

(define enter-notify-mode
  (lambda (mode)
    (case mode
      ((0) 'normal)
      ((1) 'grab)
      ((2) 'ungrab)
      (else (error "Bad mode received from xserver" mode)))))

(define enter-notify-same-screen&focus
  (lambda (byte)
    (values (not (zero? (bitwise-and byte #x01)))
	    (not (zero? (bitwise-and byte #x02))))))

(define-event enter-notify-event
  ((code               	 card8)
   (detail             	 card8     enter-notify-detail)
   (sequence-number    	 card16)
   (time               	 timestamp)
   (root               	 window)
   (event              	 window)
   (child              	 window    zero-window)
   (root-x             	 int16)
   (root-y             	 int16)
   (event-x            	 int16)
   (event-y            	 int16)
   (state              	 keybutmask)
   (mode                 card8     enter-notify-mode)
   ((same-screen focus)  card8     enter-notify-same-screen&focus)
   ))


;;;;;;;;;;;;;;
;;; Expose ;;;
;;;;;;;;;;;;;;

(define-event expose-event
  ((code            card8)
   (unused          card8)
   (sequence-number card16)
   (window          window)
   (x               card16)
   (y               card16)
   (width           card16)
   (height          card16)
   (count           card16)
   (unused          (ignore 14))
   ))


;;;;;;;;;;;;;;;;;;;;;;;;;
;;; graphics-exposure ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;

(define-event graphics-exposure-event
  ((code            card8)
   (unused          card8)
   (sequence-number card16)
   (drawable        drawable)
   (x       	    card16)
   (y       	    card16)
   (width    	    card16)
   (height   	    card16)
   (minor-opcode    card16)
   (count           card16)
   (major-opcode    card16)
   (unused          (ignore 11))
   ))


;;;;;;;;;;;;;;;;;
;;; key-press ;;;
;;;;;;;;;;;;;;;;;

(define-event key-press-event
  ((code             card8)
   (detail           keycode)
   (sequence-number  card16)
   (time             timestamp)
   (root             window)
   (event            window)
   (child            window      zero-window)
   (root-x           int16)
   (root-y           int16)
   (event-x          int16)
   (event-y          int16)
   (state            keybutmask)
   (same-screen      bool)
   (unused           (ignore 1))
   ))


;;;;;;;;;;;;;;;;;;;
;;; key-release ;;;
;;;;;;;;;;;;;;;;;;;

(define-event key-release-event
  ((code            card8)
   (detail          keycode)
   (sequence-number card16)
   (time            timestamp)
   (root            window)
   (event           window)
   (child           window        zero-window)
   (root-x          int16)
   (root-y          int16)
   (event-x         int16)
   (event-y         int16)
   (state           keybutmask)
   (same-screen     bool)
   (unused          (ignore 1))
   ))


;;;;;;;;;;;;;;;;;;;;
;;; leave-notify ;;;
;;;;;;;;;;;;;;;;;;;;

;;; alter functions really are the same as in enter-notify.

(define-event leave-notify-event
  ((code                 card8)
   (detail               card8    enter-notify-detail)
   (sequence-number      card16)
   (time                 timestamp)
   (root                 window)
   (event                window)
   (child                window   zero-window)
   (root-x               int16)
   (root-y               int16)
   (event-x              int16)
   (event-y              int16)
   (state                keybutmask)
   (mode                 card8    enter-notify-mode)
   ((same-screen focus)  card8    enter-notify-same-screen&focus)
   ))


;;;;;;;;;;;;;;;;;
;;; no-expose ;;;
;;;;;;;;;;;;;;;;;

(define-event no-expose-event
  ((code             card8)
   (unused           card8)
   (sequence-number  card16)
   (drawable         drawable)
   (minor-opcode     card16)
   (major-opcode     card8)
   (unused           (ignore 21))
   ))
