;;; frog.scm

;;; Note.  Froggy doesn't work yet.

;;;
;;; Interval-Bound Integers
;;;

;;; An interval-bound integer is one that it either wraps
;;; around or sticks when it is incremented or decremented
;;; beyond its bounds.
;;; Thus its value always lies within start (inclusive) and end (exclusive).

(define-record ibi
  min max step bound-action value)

(define change-ibi-value
  (lambda (ibi step +/-)
    (if (ibi? ibi)
	(let ((value  (ibi:value ibi))
	      (max    (ibi:max ibi))
	      (min    (ibi:min ibi))
	      (action (ibi:bound-action ibi))
	      (step^  (ibi:step ibi)))
	  (let* ((step (if (pair? step) (car step) step^))
		 (new-value (+/- value step)))
	    (set-ibi:value
	     ibi (case action
		   ((wrap) (let loop ((new-value new-value))
			     (cond ((>= new-value max) (loop (+ (- new-value max) min)))
				   ((< new-value min) (loop (+ new-value (- max min))))
				   (else new-value))))
		   ((stick) (cond ((>= new-value max) (- max 1))
				  ((< new-value min) min)
				  (else new-value)))))
	    ibi))
	(error "not an interval-bound integer" ibi))))

(define increment! (lambda (ibi . step) (change-ibi-value ibi step +)))
(define decrement! (lambda (ibi . step) (change-ibi-value ibi step -)))

(define display-ibi
  (lambda (ibi)
    (format #t "min: ~a  max: ~a  default step: ~a  bound action: ~a  value: ~a~%"
	    (ibi:min ibi)  (ibi:max ibi)
	    (ibi:step ibi) (ibi:bound-action ibi)
	    (ibi:value ibi))))

(define value
  (lambda (ibi)
    (if (ibi? ibi)
	(ibi:value ibi)
	(error "not an interval-bound integer" ibi))))

;;(define copy-ibi
;;  (lambda (ibi)
;;    (make-ibi
;;     (ibi:min ibi)  (ibi:max ibi)
;;     (ibi:step ibi) (ibi:bound-action ibi)
;;     (ibi:value ibi))))


;;;
;;; Program specific tools
;;;

;;; How do you generalize this and stick it in the xp code?

(define interpret-key
  (let ((q 21) (h 51) (j 59) (k 66) (l 75))
    (lambda (key)
      (cond ((= key h) 'left)
	    ((= key l) 'right)
	    ((= key k) 'up)
	    ((= key j) 'down)
	    ((= key q) 'quit)
	    (else (format #t "unknown key: ~a~%" key)
		  'ignore)))))


;;;
;;; The Frogster
;;;

;;; The parent thread controls frog movement.
;;; The parent thread also controls clean-up of non-log lines.
;;; Each horizontal line has a tending thread that keeps up with its x position.
;;; When the froggy moves, the parent process removes him from his old home
;;;  and installs him at his new home.  both old and new home are redrawn.

(define logs/line 5)  ; the number of log/water pairs on a log line.
(define lines     4)  ; the number of lines containing moving logs.
(define log-length 40)  ; pixels
(define vertical-quantum 20)  ; vertical quanta

(define log-period (* 2 log-length))
(define all-lines (+ 2 lines))  ; log lines and top and bottom ground lines
(define window-width (* logs/line log-length))

;;; log lines + a top line and a bottom line of dirt.
(define window-height/grid (+ 2 lines))
(define window-height (* vertical-quantum window-height/grid))


;;; the state of each thread

(define-record log-line lock froggy-position draw-func)

(define *thread-state*
  (list->vector (repeat-fn all-lines
			   (lambda () (make-log-line (make-lock) #f 'uninitialized)))))

(define thread-i
  (lambda (i)
    (vector-ref *thread-state* i)))

(define thread-for-each
  (lambda (f)
    (vector-for-each f *thread-state*)))

(define display-log-line
  (lambda (log-line)
    (format #t "log-line::  lock: ~a  froggy-position: ~a  draw-func: ~a~%"
	    (log-line:lock log-line)
	    (log-line:froggy-position log-line)
	    (log-line:draw-func log-line))))

(define set-froggy-position
  (lambda (log new-position)
    (with-lock (log-line:lock log)
      (lambda ()
	(set-log-line:froggy-position log new-position)))))


;;; X stuff that gets in the way

(define *connection* 'not-defined-yet)
(define *colormap*   'not-defined-yet)
(define *window*     'not-defined-yet)
(define *gc*         'not-defined-yet)
(define *blue*       'not-defined-yet)
(define *brown*      'not-defined-yet)
(define *green*      'not-defined-yet)


;;; Use of modulo will bring the problem down to one log/water pair.
;;; We now have a view like:  LLLLLLWWWWWW
;;; with a position somewhere in that range.
;;; If we subtract the length of one log, the position will drift off
;;; the left edge and be negative or zero if we were on the log.

(define point-on-log?
  (lambda (thread-number x)
    (let ((line (thread-i thread-number)))
      (with-lock (log-line:lock line)
	(lambda ()
	  (let ((offset (modulo x log-length)))
	    (not (positive? (- offset log-length)))))))))


(define draw-froggy
  (lambda (x y)
    (change-gc *connection* *gc* `((foreground ,*green*)))
    (poly-fill-rectangle *connection* *window* *gc*
			 `((,x ,(* y vertical-quantum) ,vertical-quantum ,vertical-quantum)))
    ))

(define log-drawer
  (lambda (y)
    (lambda (x)
      (let* ((line-y (* y vertical-quantum))
	     (rec-tail (list line-y log-length vertical-quantum))
	     (rectangles (let loop ((n (+ logs/line 1)) (x x) (recs '()))
			   (if (zero? n)
			       recs
			       (loop (- n 1) (+ x log-period)
				     (cons (cons x rec-tail) recs))))))
	(change-gc *connection* *gc* `((foreground ,*brown*)))
	(poly-fill-rectangle *connection* *window* *gc* rectangles)
	))))

(define water-drawer
  (lambda (y)
    (lambda (x)
      (let* ((line-y (* y vertical-quantum))
	     (rec-tail (list line-y log-length vertical-quantum))
	     (rectangles (let loop ((n (+ logs/line 1)) (x x) (recs '()))
			   (if (zero? n)
			       recs
			       (loop (- n 1) (+ x log-period)
				     (cons (cons x rec-tail) recs))))))
	(change-gc *connection* *gc* `((foreground ,*blue*)))
	(poly-fill-rectangle *connection* *window* *gc* rectangles)
	))))

(define ground-drawer
  (lambda (y)
    (let ((line (thread-i y)))
      (lambda ()
	(change-gc *connection* *gc* `((foreground ,*brown*)))
	(poly-fill-rectangle *connection* *window* *gc*
			     `((0 ,(* y vertical-quantum) ,window-width ,vertical-quantum)))
	(cond ((log-line:froggy-position line) =>
	       (lambda (froggy-x)
		 (draw-froggy froggy-x y))))
	))))

(define ground-maker
  (lambda (line-number wait)
    (lambda ()
      (let ((line (thread-i line-number))
	    (draw-line (ground-drawer line-number)))
	(set-log-line:draw-func line draw-line)
	(draw-line)
	))))


;;; We check to see if froggy is sitting on us.
;;; If he is, we increment him too.

(define line-drawer
  (lambda (line-number start-x)
    (let ((x-ibi (make-ibi 0 log-period 1 'wrap start-x))
	  (log-line   (thread-i     line-number))
	  (draw-logs  (log-drawer   line-number))
	  (draw-water (water-drawer line-number)))
      (lambda ()
	(increment! x-ibi)
	(let ((x (value x-ibi)))
	  (draw-logs  (- x log-period))
	  (draw-water (- x log-length)))
	(cond ((log-line:froggy-position log-line) =>
	       (lambda (old-x)
		 (let ((new-x (+ old-x 1)))
		   (set-froggy-position log-line new-x)
		   (draw-froggy new-x line-number))))))
      )))

(define log-line-maker
  (lambda (line-number wait start-x)
    (let ((log-line (thread-i line-number)))
      (lambda ()
	(let ((draw-line (line-drawer line-number start-x)))
	  (set-log-line:draw-func log-line draw-line)
	  (infinite-loop (draw-line)
			 ;;(relinquish-timeslice)
			 (thread-sleep wait)
			 ))))))

(define move-froggy
  (let ((y-ibi (make-ibi 0 window-height/grid 1 'stick (- window-height/grid 1))))
    (lambda (dir)
      (case dir
	((down)  (let* ((old-y (value y-ibi))
			(new-y (value (increment! y-ibi)))
			(old-home (thread-i old-y))
			(new-home (thread-i new-y))
			(x-position (log-line:froggy-position old-home)))
		   (when (< old-y new-y)
		     (set-froggy-position old-home #f)
		     (set-froggy-position new-home x-position)
		     ((log-line:draw-func old-home))
		     ((log-line:draw-func new-home)))))
	((up)    (let* ((old-y (value y-ibi))
			(new-y (value (decrement! y-ibi)))
			(old-home (thread-i old-y))
			(new-home (thread-i new-y))
			(x-position (log-line:froggy-position old-home)))
		   (when (> old-y new-y)
		     (set-froggy-position old-home #f)
		     (set-froggy-position new-home x-position)
		     ((log-line:draw-func old-home))
		     ((log-line:draw-func new-home)))))
	((left)  (let* ((home (thread-i (value y-ibi)))
			(old-position (log-line:froggy-position home))
			(new-position (- old-position vertical-quantum)))
		   (set-froggy-position home new-position)
		   ((log-line:draw-func home))))
	((right) (let* ((home (thread-i (value y-ibi)))
			(old-position (log-line:froggy-position home))
			(new-position (+ old-position vertical-quantum)))
		   (set-froggy-position home new-position)
		   ((log-line:draw-func home))))
	(else (error "weird direction" dir))))))


(define froggy
  (lambda ()
    (with-connection
     (lambda (connection)
       (let* ((parent (root-window connection))
	      (depth  (root-depth connection))
	      (white  (white-pixel connection))
	      (black  (black-pixel connection))
	      (window (create-window connection parent 0 0 window-width window-height
				     1 depth 'copy-from-parent 'copy-from-parent
				     `((event-mask (exposure button-press key-press))
				       (background-pixel ,white))))
	      (gc (create-gc connection window `((foreground ,black)
						 (background ,white)
						 (graphics-exposures #f) )))
	      (colormap (default-colormap connection))
	      (brown  (alloc-named-color-reply:pixel
		       (alloc-named-color connection colormap "brown")))
	      (blue   (alloc-named-color-reply:pixel
		       (alloc-named-color connection colormap "blue")))
	      (green  (alloc-named-color-reply:pixel
		       (alloc-named-color connection colormap "green")))
	      (window-lock (make-lock))
	      ;; for double buffering
	      (before-pixmap (create-pixmap connection window depth window-width window-height))
	      (after-pixmap  (create-pixmap connection window depth window-width window-height)))
	 (format #t "Map Window~%")
	 (map-window connection window)
	 (set! *connection* connection)
	 (set! *window*     window)
	 (set! *colormap*   colormap)
	 (set! *gc*         gc)
	 (set! *blue*       blue)
	 (set! *brown*      brown)
	 (set! *green*      green)
	 (with-multitasking
	  (lambda ()
	    (call-with-current-continuation
	     (lambda (finish)
	       (let* ((max-random (expt 2 28))
		      (max-wait 0)
		      (random (make-random 0))
		      (wait-maker (lambda ()
				    (round (* (/ (random) max-random) max-wait))))
		      (first-x (lambda () (round (* (/ (random) max-random) log-period))))
		      (logs (let loop ((n lines)
				       (logs (list (ground-maker (+ lines 1) (wait-maker)))))
			      (if (zero? n)
				  (cons (ground-maker 0 (wait-maker)) logs)
				  (loop (- n 1) (cons (log-line-maker n (wait-maker) (first-x))
						      logs)))))
		      (logs (map spawn logs)))
		 (set-log-line:froggy-position (thread-i (+ lines 1))
					       (quotient window-width 2))
		 (let loop ((event (read-event connection)))
		   (cond ((or (expose-event? event) (graphics-exposure-event? event))
			  (thread-for-each (lambda (log) ((log-line:draw-func log)))))
			 ((key-press-event? event)
			  (let ((action (interpret-key (key-press-event:detail event))))
			    (case action
			      ((up down left right) (move-froggy action))
			      ((quit)  (for-each kill-thread logs) (finish #t))
			      ((ignore) #f)
			      (else    (error "odd action" action)))))
			 (else (format #t "strange event~%")))
		   (loop (read-event connection)))))))))))))
