;;; structure-datatypes.scm


;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Global ID Numbers ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;

;;; During Connection Setup, the server gives a resource-id-mask and resource-id-base.
;;; New ids for windows, pixmaps, etc. are base on these values.
;;; The documentation isn't really clear to me how the ids are derived.

(define *id-base* 0)
(define *id-mask* 0)
(define *id* 0)

(define initialize-ids
  (lambda (base mask)
    (begin
      (set! *id-base* base)
      (set! *id-mask* mask)
      (set! *id* 1))))

(define new-id
  (lambda ()
    (begin (set! *id* (+ *id* 1))
	   (if (> *id* *id-mask*)
	       (error "id overflow"))
	   (bitwise-ior *id-base* *id*))))


;;;;;;;;;;;;;;;;;;;;;
;;; Pixmap-Format ;;;
;;;;;;;;;;;;;;;;;;;;;

(define-record pixmap-format
  depth                            ; card8
  bits/pixel                       ; card8 (1, 4, 8, 16, 24, 32)
  scanline-pad)                    ; card8 (8, 16, 32)

(define read-pixmap-format
  (lambda (xserver)
    (let ((depth        (read-card8 xserver))
	  (bits/pixel   (read-card8 xserver))
	  (scanline-pad (read-card8 xserver))
	  (unused       (ignore 5 xserver)))
      (make-pixmap-format depth bits/pixel scanline-pad))))

(define display-pixmap-format
  (lambda (pixmap-format)
    (format #t (string-append
		"PIXMAP-FORMAT~%"
		"  depth:        ~a~%"
		"  bits/pixel:   ~a~%"
		"  scanline-pad: ~a~%")
	    (pixmap-format:depth pixmap-format)
	    (pixmap-format:bits/pixel pixmap-format)
	    (pixmap-format:scanline-pad pixmap-format))))


;;;;;;;;;;;;;;;;;;;
;;; Visual-Type ;;;
;;;;;;;;;;;;;;;;;;;

(define-record visual-type
  visual-id                    ; visual-id (card32)
  class                        ; class (card8) (0-5)
  bits/rgb-value               ; card8
  colormap-entries             ; card16
  red-mask                     ; card32
  green-mask                   ; card32
  blue-mask)                   ; card32

(define visual-type-class-alist `((static-gray  . 0)
				  (gray-scale   . 1)
				  (static-color . 2)
				  (pseudo-color . 3)
				  (true-color   . 4)
				  (direct-color . 5)))


;;; (visual-type-class->symbol visual-type-class) => symbol

(define visual-type-class->symbol
  (value->symbol visual-type-class-alist
		 (lambda (val)
		   (error "input not class" val))))


;;; (symbol->visual-type-class symbol) => visual-type-class

(define symbol->visual-type-class
  (symbol->value visual-type-class-alist
		 (lambda (sym)
		   (error "unknown symbol" sym))))


;;; (read-visual-type xserver) => unspecified

(define read-visual-type
  (lambda (xserver)
    (let* ((visual-id   (read-visual-id xserver))
	   (class       (visual-type-class->symbol (read-card8 xserver)))
	   (bits/rgb-value   (read-card8 xserver))
	   (colormap-entries (read-card16 xserver))
	   (red-mask    (read-card32 xserver))
	   (green-mask  (read-card32 xserver))
	   (blue-mask   (read-card32 xserver))
	   (unused      (ignore 4 xserver)))
      (make-visual-type visual-id class bits/rgb-value colormap-entries
			red-mask green-mask blue-mask))))

(define display-visual-type
  (lambda (visual-type)
    (format #t (string-append
		"VISUAL-TYPE~%"
		"  visual-id:        ~a~%"
		"  class:            ~a~%"
		"  bits/rgb-value:   ~a~%"
		"  colormap-entries: ~a~%"
		"  red-mask:         ~a~%"
		"  green-mask:       ~a~%"
		"  blue-mask:        ~a~%")
	    (visual-type:visual-id visual-type)
	    (visual-type:class visual-type)
	    (visual-type:bits/rgb-value visual-type)
	    (visual-type:colormap-entries visual-type)
	    (visual-type:red-mask visual-type)
	    (visual-type:green-mask visual-type)
	    (visual-type:blue-mask visual-type))))


;;;;;;;;;;;;;;
;;; Depths ;;;
;;;;;;;;;;;;;;

(define-record depth
  depth-value	           ; card8
  visuals-list)            ; (visual-type)

(define read-depth
  (lambda (xserver)
    (let* ((depth-value            (read-card8 xserver))
	   (unused                 (ignore 1 xserver))
	   (number-of-visual-types (read-card16 xserver))
	   (unused                 (ignore 4 xserver))
	   (visuals-list           (gather number-of-visual-types
					   (lambda () (read-visual-type xserver)))))
      (make-depth depth-value visuals-list))))

(define display-depth
  (lambda (depth)
    (format #t (string-append
		"DEPTH~%"
		"  depth-value: ~a~%")
	    (depth:depth-value depth))
    (for-each display-visual-type (depth:visuals-list depth))))


;;;;;;;;;;;;;;
;;; Screen ;;;
;;;;;;;;;;;;;;

(define-record screen
  root                   ; window
  default-colormap       ; colormap
  white-pixel            ; card32
  black-pixel            ; card32
  current-input-masks	 ; set-of-event (card32)
  width-in-pixels        ; card16
  height-in-pixels       ; card16
  width-in-millimeters   ; card16
  height-in-millimeters  ; card16
  min-installed-maps     ; card16
  max-installed-maps     ; card16
  root-visual            ; visual-id (card32)
  backing-stores         ; card8 (0, 1, 2)
  save-unders            ; bool
  root-depth             ; card8
  depths-list)           ; (depth)

(define read-screen
  (lambda (xserver)
    (let* ((root                  (read-window xserver))
	   (default-colormap      (read-colormap xserver))
	   (white-pixel           (read-card32 xserver))
	   (black-pixel           (read-card32 xserver))
	   (current-input-masks   (read-set-of-event xserver))
	   (width-in-pixels       (read-card16 xserver))
	   (height-in-pixels      (read-card16 xserver))
	   (width-in-millimeters  (read-card16 xserver))
	   (height-in-millimeters (read-card16 xserver))
	   (min-installed-maps    (read-card16 xserver))
	   (max-installed-maps    (read-card16 xserver))
	   (root-visual           (read-visual-id xserver))
	   (backing-stores        (let ((val (read-card8 xserver)))
				    (case val
				      ((0) 'never)
				      ((1) 'when-mapped)
				      ((2) 'always)
				      (else (error "odd value for backing-stores" val)))))
	   (save-unders           (read-bool xserver))
	   (root-depth            (read-card8 xserver))
	   (number-of-depths      (read-card8 xserver))
	   (depths-list           (gather number-of-depths (lambda () (read-depth xserver)))))
      (make-screen root default-colormap white-pixel black-pixel
		   current-input-masks width-in-pixels height-in-pixels
		   width-in-millimeters height-in-millimeters
		   min-installed-maps max-installed-maps root-visual
		   backing-stores save-unders root-depth depths-list))))

(define display-screen
  (lambda (screen)
    (format #t (string-append
		"SCREEN~%"
		"  root:                  ~a~%"
		"  default-colormap:      ~a~%"
		"  white-pixel:           ~a~%"
		"  black-pixel:           ~a~%"
		"  current-input-masks:   ~a~%"
		"  width-in-pixels:       ~a~%"
		"  height-in-pixels:      ~a~%"
		"  width-in-millimeters:  ~a~%"
		"  height-in-millimeters: ~a~%"
		"  min-installed-maps:    ~a~%"
		"  max-installed-maps:    ~a~%"
		"  root-visual:           ~a~%"
		"  backing-stores:        ~a~%"
		"  save-unders:           ~a~%"
		"  root-depth:            ~a~%")
	    (screen:root screen)
	    (screen:default-colormap screen)
	    (screen:white-pixel screen)
	    (screen:black-pixel screen)
	    (screen:current-input-masks screen)
	    (screen:width-in-pixels screen)
	    (screen:height-in-pixels screen)
	    (screen:width-in-millimeters screen)
	    (screen:height-in-millimeters screen)
	    (screen:min-installed-maps screen)
	    (screen:max-installed-maps screen)
	    (screen:root-visual screen)
	    (screen:backing-stores screen)
	    (screen:save-unders screen)
	    (screen:root-depth screen))
    (for-each display-depth (screen:depths-list screen))))


;;;;;;;;;;;;;;;;;;
;;; Connection ;;;
;;;;;;;;;;;;;;;;;;

(define-record connection
  xserver                         ; xserver
  protocol-major-version          ; card16
  protocol-minor-version          ; card16
  release-number                  ; card32
  resource-id-base                ; card32
  resource-id-mask                ; card32
  motion-buffer-size              ; card32
  maximum-request-length          ; card16
  image-byte-order                ; card8 (0, 1)
  bitmap-format-bit-order         ; card8 (0, 1)
  bitmap-format-scanline-unit     ; card8
  bitmap-format-scanline-pad      ; card8
  min-keycode                     ; keycode (card8)
  max-keycode                     ; keycode (card8)
  vendor                          ; string8
  pixmap-formats-list             ; (pixmap-format)
  root-screens-vector)            ; #(screen)

(define connection-xserver connection:xserver)

(define read-connection
  (lambda (xserver)
    (let* ((unused                    (ignore 1 xserver))
	   (protocol-major-version    (read-card16 xserver))
	   (protocol-minor-version    (read-card16 xserver))
	   (length-of-add-data        (read-card16 xserver))
	   (release-number            (read-card32 xserver))
	   (resource-id-base          (read-card32 xserver))
	   (resource-id-mask          (read-card32 xserver))
	   (motion-buffer-size        (read-card32 xserver))
	   (length-of-vendor          (read-card16 xserver))
	   (maximum-request-length    (read-card16 xserver))
	   (number-of-root-screens    (read-card8 xserver))
	   (number-of-pixmap-formats  (read-card8 xserver))
	   (image-byte-order  (let ((val (read-card8 xserver)))
				(case val
				  ((0) 'lsb-first)
				  ((1) 'msb-first)
				  (else (error "odd value returned for image-byte-order" val)))))
	   (bitmap-format-bit-order     (read-card8 xserver))
	   (bitmap-format-scanline-unit (read-card8 xserver))
	   (bitmap-format-scanline-pad  (read-card8 xserver))
	   (min-keycode                 (read-keycode xserver))
	   (max-keycode                 (read-keycode xserver))
	   (unused               (ignore 4 xserver))
	   (vendor               (string8->string (read-string8 length-of-vendor xserver)))
	   (unused               (ignore (pad-length length-of-vendor) xserver))
	   (pixmap-formats-list  (gather number-of-pixmap-formats
					 (lambda () (read-pixmap-format xserver)) ))
	   (root-screens-vector  (list->vector
				  (gather number-of-root-screens
					  (lambda () (read-screen xserver))))))
      (initialize-ids resource-id-base resource-id-mask)
      (make-connection
       xserver protocol-major-version protocol-minor-version
       release-number resource-id-base resource-id-mask
       motion-buffer-size maximum-request-length
       image-byte-order bitmap-format-bit-order
       bitmap-format-scanline-unit
       bitmap-format-scanline-pad min-keycode max-keycode
       vendor pixmap-formats-list root-screens-vector))))

(define display-connection
  (lambda (connection)
    (format #t (string-append
		"CONNEXTION~%"
		"  protocol-major-version: ~a~%"
		"  protocol-minor-version: ~a~%"
		"  release-number: ~a~%"
		"  resource-id-base: ~a~%"
		"  resource-id-mask: ~a~%"
		"  motion-buffer-size: ~a~%"
		"  maximum-request-length: ~a~%"
		"  image-byte-order: ~a~%"
		"  bitmap-format-bit-order: ~a~%"
		"  bitmap-format-scanline-unit: ~a~%"
		"  bitmap-format-scanline-pad: ~a~%"
		"  min-keycode: ~a~%"
		"  max-keycode: ~a~%"
		"  vendor: ~a~%")
	    (connection:protocol-major-version connection)
	    (connection:protocol-minor-version connection)
	    (connection:release-number connection)
	    (connection:resource-id-base connection)
	    (connection:resource-id-mask connection)
	    (connection:motion-buffer-size connection)
	    (connection:maximum-request-length connection)
	    (connection:image-byte-order connection)
	    (connection:bitmap-format-bit-order connection)
	    (connection:bitmap-format-scanline-unit connection)
	    (connection:bitmap-format-scanline-pad connection)
	    (connection:min-keycode connection)
	    (connection:max-keycode connection)
	    (connection:vendor connection))
    (for-each display-pixmap-format (connection:pixmap-formats-list connection))
    (for-each display-screen (vector->list
			      (connection:root-screens-vector connection)))))
